@@ -277,9 +277,7 @@ char* nav_find_main_title(const char *root)
         mpls_free(pl_list[ii]);
     }
     if (count > 0) {
-        char *str = (char*)malloc(strlen(longest) + 1);
-        strcpy(str, longest);
-        return str;
+        return strdup(longest);
     } else {
         return NULL;
     }
@@ -466,8 +464,7 @@ NAV_TITLE* nav_title_open(const char *root, const char *playlist, unsigned angle
     if (title == NULL) {
         return NULL;
     }
-    title->root = (char*)malloc(strlen(root) + 1);
-    strcpy(title->root, root);
+    title->root = strdup(root);
     strncpy(title->name, playlist, 11);
     title->name[10] = '\0';
     path = str_printf("%s" DIR_SEP "BDMV" DIR_SEP "PLAYLIST" DIR_SEP "%s",
