@@ -34,6 +34,7 @@
 #include "bdnav/index_parse.h"
 #include "bdnav/meta_parse.h"
 #include "bdnav/clpi_parse.h"
+#include "bdnav/sound_parse.h"
 #include "hdmv/hdmv_vm.h"
 #include "decoders/graphics_controller.h"
 #include "file/file.h"
@@ -46,6 +47,19 @@
 #include <inttypes.h>
 #include <string.h>
 
+#ifdef HAVE_MNTENT_H
+#include <sys/stat.h>
+#include <mntent.h>
+#endif
+
+#ifdef __APPLE__
+#define _DARWIN_C_SOURCE
+#include <sys/stat.h>
+#include <sys/param.h>
+#include <sys/ucred.h>
+#include <sys/mount.h>
+#endif
+
 typedef int     (*fptr_int)();
 typedef int32_t (*fptr_int32)();
 typedef void*   (*fptr_p_void)();
@@ -141,6 +155,8 @@ struct bluray {
 
     /* graphics */
     GRAPHICS_CONTROLLER *graphics_controller;
+    SOUND_DATA          *sound_effects;
+    uint32_t             gc_status;
 };
 
 #define DL_CALL(lib,func,param,...)             \
@@ -549,7 +565,7 @@ static int _run_gc(BLURAY *bd, gc_ctrl_e msg, uint32_t param)
     int result = -1;
 
     if (bd && bd->graphics_controller && bd->hdmv_vm) {
-        GC_NAV_CMDS cmds = {-1, NULL, -1};
+        GC_NAV_CMDS cmds = {-1, NULL, -1, 0};
 
         result = gc_run(bd->graphics_controller, msg, param, &cmds);
 
@@ -557,6 +573,30 @@ static int _run_gc(BLURAY *bd, gc_ctrl_e msg, uint32_t param)
             hdmv_vm_set_object(bd->hdmv_vm, cmds.num_nav_cmds, cmds.nav_cmds);
             bd->hdmv_suspended = !hdmv_vm_running(bd->hdmv_vm);
         }
+
+        if (cmds.status != bd->gc_status) {
+            uint32_t changed_flags = cmds.status ^ bd->gc_status;
+            bd->gc_status = cmds.status;
+            if (changed_flags & GC_STATUS_MENU_OPEN) {
+                _queue_event(bd, (BD_EVENT){BD_EVENT_MENU, !!(bd->gc_status & GC_STATUS_MENU_OPEN)});
+            }
+            if (changed_flags & GC_STATUS_POPUP) {
+                _queue_event(bd, (BD_EVENT){BD_EVENT_POPUP, !!(bd->gc_status & GC_STATUS_POPUP)});
+            }
+        }
+
+        if (cmds.sound_id_ref >= 0 && cmds.sound_id_ref < 0xff) {
+            _queue_event(bd, (BD_EVENT){BD_EVENT_SOUND_EFFECT, cmds.sound_id_ref});
+        }
+
+    } else {
+        if (bd->gc_status & GC_STATUS_MENU_OPEN) {
+            _queue_event(bd, (BD_EVENT){BD_EVENT_MENU, 0});
+        }
+        if (bd->gc_status & GC_STATUS_POPUP) {
+            _queue_event(bd, (BD_EVENT){BD_EVENT_POPUP, 0});
+        }
+        bd->gc_status = GC_STATUS_NONE;
     }
 
     return result;
@@ -626,6 +666,12 @@ static int _libaacs_load(BLURAY *bd)
         if (bd->libaacs_open && bd->libaacs_decrypt_unit) {
             BD_DEBUG(DBG_BLURAY, "Loaded libaacs (%p)\n", bd->h_libaacs);
             bd->disc_info.libaacs_detected = 1;
+
+            if (file_open != file_open_default()) {
+                BD_DEBUG(DBG_BLURAY, "Registering libaacs filesystem handler %p (%p)\n", file_open, bd->h_libaacs);
+                DL_CALL(bd->h_libaacs, aacs_register_file, file_open);
+            }
+
             return 1;
 
         } else {
@@ -822,8 +868,8 @@ static int _index_open(BLURAY *bd)
 
 static int _meta_open(BLURAY *bd)
 {
-    if (!bd->meta){
-      bd->meta = meta_parse(bd->device_path);
+    if (!bd->meta) {
+        bd->meta = meta_parse(bd->device_path);
     }
 
     return !!bd->meta;
@@ -874,40 +920,109 @@ static void _fill_disc_info(BLURAY *bd)
     }
 }
 
+#ifdef HAVE_MNTENT_H
+/*
+ * Replace device node (/dev/sr0) by mount point
+ * Implemented on Linux and MacOSX
+ */
+static void get_mount_point(BLURAY *bd)
+{
+    struct stat st;
+    if (stat (bd->device_path, &st) )
+        return;
+
+    /* If it's a directory, all is good */
+    if (S_ISDIR(st.st_mode))
+        return;
+
+    FILE *f = setmntent ("/proc/self/mounts", "r");
+    if (!f)
+        return;
+
+    struct mntent* m;
+#ifdef HAVE_GETMNTENT_R
+    struct mntent mbuf;
+    char buf [8192];
+    while ((m = getmntent_r (f, &mbuf, buf, sizeof(buf))) != NULL) {
+#else
+    while ((m = getmntent (f)) != NULL) {
+#endif
+        if (!strcmp (m->mnt_fsname, bd->device_path)) {
+            free(bd->device_path);
+            bd->device_path = strdup (m->mnt_dir);
+            break;
+        }
+    }
+    endmntent (f);
+}
+#endif
+#ifdef __APPLE__
+static void get_mount_point(BLURAY *bd)
+{
+    struct stat st;
+    if (stat (bd->device_path, &st) )
+        return;
+
+    /* If it's a directory, all is good */
+    if (S_ISDIR(st.st_mode))
+        return;
+
+    struct statfs mbuf[128];
+    int fs_count;
+
+    if ( (fs_count = getfsstat (NULL, 0, MNT_NOWAIT)) == -1 ) {
+        return;
+    }
+
+    getfsstat (mbuf, fs_count * sizeof(mbuf[0]), MNT_NOWAIT);
+
+    for ( int i = 0; i < fs_count; ++i) {
+        if (!strcmp (mbuf[i].f_mntfromname, bd->device_path)) {
+            free(bd->device_path);
+            bd->device_path = strdup (mbuf[i].f_mntonname);
+        }
+    }
+}
+#endif
+
 /*
  * open / close
  */
 
 BLURAY *bd_open(const char* device_path, const char* keyfile_path)
 {
-    BLURAY *bd = calloc(1, sizeof(BLURAY));
-
     BD_DEBUG(DBG_BLURAY, "libbluray version "BLURAY_VERSION_STRING"\n");
 
-    if (device_path) {
-
-        bd->device_path = (char*)malloc(strlen(device_path) + 1);
-        strcpy(bd->device_path, device_path);
-
-        _libaacs_open(bd, keyfile_path);
-
-        _libbdplus_open(bd, keyfile_path);
-
-        _index_open(bd);
-
-        bd->meta = NULL;
-
-        bd->regs = bd_registers_init();
-
-        _fill_disc_info(bd);
-
-        BD_DEBUG(DBG_BLURAY, "BLURAY initialized! (%p)\n", bd);
-    } else {
-        X_FREE(bd);
-
+    if (!device_path) {
         BD_DEBUG(DBG_BLURAY | DBG_CRIT, "No device path provided!\n");
+        return NULL;
     }
 
+    BLURAY *bd = calloc(1, sizeof(BLURAY));
+
+    if (!bd) {
+        BD_DEBUG(DBG_BLURAY | DBG_CRIT, "Can't allocate memory\n");
+        return NULL;
+    }
+
+    bd->device_path = strdup(device_path);
+
+#if (defined HAVE_MNTENT_H || defined __APPLE__)
+    get_mount_point(bd);
+#endif
+
+    _libaacs_open(bd, keyfile_path);
+
+    _libbdplus_open(bd, keyfile_path);
+
+    _index_open(bd);
+
+    bd->regs = bd_registers_init();
+
+    _fill_disc_info(bd);
+
+    BD_DEBUG(DBG_BLURAY, "BLURAY initialized! (%p)\n", bd);
+
     return bd;
 }
 
@@ -933,6 +1048,7 @@ void bd_close(BLURAY *bd)
 
     gc_free(&bd->graphics_controller);
     indx_free(&bd->index);
+    sound_free(&bd->sound_effects);
     bd_registers_free(bd->regs);
 
     _free_event_queue(bd);
@@ -1804,6 +1920,7 @@ static void _process_psr_restore_event(BLURAY *bd, BD_PSR_EVENT *ev)
         case PSR_TIME:
             bd_seek_time(bd, ((int64_t)ev->new_val) << 1);
             _init_ig_stream(bd);
+            _run_gc(bd, GC_CTRL_INIT_MENU, 0);
             return;
 
         case PSR_SELECTED_BUTTON_ID:
@@ -2292,11 +2409,41 @@ void bd_register_overlay_proc(BLURAY *bd, void *handle, bd_overlay_proc_f func)
     }
 }
 
+int bd_get_sound_effect(BLURAY *bd, unsigned sound_id, BLURAY_SOUND_EFFECT *effect)
+{
+    if (!bd || !effect) {
+        return -1;
+    }
+
+    if (!bd->sound_effects) {
+
+        char *file = str_printf("%s/BDMV/AUXDATA/sound.bdmv", bd->device_path);
+        bd->sound_effects = sound_parse(file);
+        X_FREE(file);
+
+        if (!bd->sound_effects) {
+            return -1;
+        }
+    }
+
+    if (sound_id < bd->sound_effects->num_sounds) {
+        SOUND_OBJECT *o = &bd->sound_effects->sounds[sound_id];
+
+        effect->num_channels = o->num_channels;
+        effect->num_frames   = o->num_frames;
+        effect->samples      = (const int16_t *)o->samples;
+
+        return 1;
+    }
+
+    return 0;
+}
+
 /*
  *
  */
 
-struct meta_dl *bd_get_meta(BLURAY *bd)
+const struct meta_dl *bd_get_meta(BLURAY *bd)
 {
     if (!bd) {
         return NULL;
