@@ -43,6 +43,7 @@
 typedef struct {
     uint16_t enabled_button;  /* enabled button id */
     uint16_t x, y, w, h;      /* button rect on overlay plane (if drawn) */
+    int      visible_object_id; /* id of currently visible object */
     int      animate_indx;    /* currently showing object index of animated button, < 0 for static buttons */
 } BOG_DATA;
 
@@ -57,8 +58,10 @@ struct graphics_controller_s {
     void          (*overlay_proc)(void *, const struct bd_overlay_s * const);
 
     /* state */
+    unsigned        ig_open;
     unsigned        ig_drawn;
     unsigned        ig_dirty;
+    unsigned        pg_open;
     unsigned        pg_drawn;
     unsigned        pg_dirty;
     unsigned        popup_visible;
@@ -199,6 +202,10 @@ static BD_PG_OBJECT *_find_object_for_button(PG_DISPLAY_SET *s,
         }
     }
 
+    if (!repeat && object_id_end < 0xfffe) {
+        object_id = object_id_end;
+    }
+
     object = _find_object(s, object_id);
 
     return object;
@@ -208,6 +215,14 @@ static BD_PG_OBJECT *_find_object_for_button(PG_DISPLAY_SET *s,
  * util
  */
 
+static int _areas_overlap(BOG_DATA *a, BOG_DATA *b)
+{
+    return !(a->x + a->w <= b->x        ||
+             a->x        >= b->x + b->w ||
+             a->y + a->h <= b->y        ||
+             a->y        >= b->y + b->h);
+}
+
 static int _is_button_enabled(GRAPHICS_CONTROLLER *gc, BD_IG_PAGE *page, unsigned button_id)
 {
     unsigned ii;
@@ -342,6 +357,7 @@ static void _reset_page_state(GRAPHICS_CONTROLLER *gc)
     for (ii = 0; ii < page->num_bogs; ii++) {
         gc->bog_data[ii].enabled_button = page->bog[ii].default_valid_button_id_ref;
         gc->bog_data[ii].animate_indx   = 0;
+        gc->bog_data[ii].visible_object_id = -1;
     }
 }
 
@@ -362,6 +378,12 @@ static void _open_osd(GRAPHICS_CONTROLLER *gc, int plane,
         };
 
         gc->overlay_proc(gc->overlay_proc_handle, &ov);
+
+        if (plane == BD_OVERLAY_IG) {
+            gc->ig_open = 1;
+        } else {
+            gc->pg_open = 1;
+        }
     }
 }
 
@@ -376,6 +398,14 @@ static void _close_osd(GRAPHICS_CONTROLLER *gc, int plane)
 
         gc->overlay_proc(gc->overlay_proc_handle, &ov);
     }
+
+    if (plane == BD_OVERLAY_IG) {
+        gc->ig_open = 0;
+        gc->ig_drawn = 0;
+    } else {
+        gc->pg_open = 0;
+        gc->pg_drawn = 0;
+    }
 }
 
 static void _flush_osd(GRAPHICS_CONTROLLER *gc, int plane, int64_t pts)
@@ -437,6 +467,7 @@ static void _clear_bog_area(GRAPHICS_CONTROLLER *gc, BOG_DATA *bog_data)
         _clear_osd_area(gc, BD_OVERLAY_IG, bog_data->x, bog_data->y, bog_data->w, bog_data->h);
 
         bog_data->x = bog_data->y = bog_data->w = bog_data->h = 0;
+        bog_data->visible_object_id = -1;
 
         gc->ig_dirty = 1;
     }
@@ -478,7 +509,9 @@ static void _select_button(GRAPHICS_CONTROLLER *gc, uint32_t button_id)
 static void _select_page(GRAPHICS_CONTROLLER *gc, uint16_t page_id)
 {
     bd_psr_write(gc->regs, PSR_MENU_PAGE_ID, page_id);
-    _clear_osd(gc, BD_OVERLAY_IG);
+    if (gc->ig_open) {
+        _clear_osd(gc, BD_OVERLAY_IG);
+    }
     _reset_page_state(gc);
 
     uint16_t button_id = _find_selected_button_id(gc);
@@ -489,8 +522,12 @@ static void _select_page(GRAPHICS_CONTROLLER *gc, uint16_t page_id)
 
 static void _gc_reset(GRAPHICS_CONTROLLER *gc)
 {
-    _close_osd(gc, BD_OVERLAY_PG);
-    _close_osd(gc, BD_OVERLAY_IG);
+    if (gc->pg_open) {
+        _close_osd(gc, BD_OVERLAY_PG);
+    }
+    if (gc->ig_open) {
+        _close_osd(gc, BD_OVERLAY_IG);
+    }
 
     gc->popup_visible = 0;
     gc->valid_mouse_position = 0;
@@ -655,6 +692,43 @@ static void _render_button(GRAPHICS_CONTROLLER *gc, BD_IG_BUTTON *button, BD_PG_
         return;
     }
 
+    /* object already rendered ? */
+    if (bog_data->visible_object_id == object->id &&
+        bog_data->x == button->x_pos && bog_data->y == button->y_pos &&
+        bog_data->w == object->width && bog_data->h == object->height) {
+
+        GC_TRACE("skipping already rendered button #%d (object #%d at %d,%d %dx%d)\n",
+                 button->id, object->id,
+                 button->x_pos, button->y_pos, object->width, object->height);
+
+        return;
+    }
+
+    /* new object is smaller than already drawn one ? -> need to render background */
+    if (bog_data->w > object->width ||
+        bog_data->h > object->height) {
+
+        /* make sure we won't wipe other buttons */
+        unsigned ii, skip = 0;
+        for (ii = 0; &gc->bog_data[ii] != bog_data; ii++) {
+            if (_areas_overlap(bog_data, &bog_data[ii]))
+                skip = 1;
+            /* FIXME: clean non-overlapping area */
+        }
+
+        GC_TRACE("object size changed, %sclearing background at %d,%d %dx%d\n",
+                 skip ? " ** NOT ** " : "",
+                 bog_data->x, bog_data->y, bog_data->w, bog_data->h);
+
+        if (!skip) {
+            _clear_bog_area(gc, bog_data);
+        }
+    }
+
+    GC_TRACE("render button #%d using object #%d at %d,%d %dx%d\n",
+             button->id, object->id,
+             button->x_pos, button->y_pos, object->width, object->height);
+
     _render_object(gc, -1, BD_OVERLAY_IG,
                    button->x_pos, button->y_pos,
                    object, palette);
@@ -663,12 +737,13 @@ static void _render_button(GRAPHICS_CONTROLLER *gc, BD_IG_BUTTON *button, BD_PG_
     bog_data->y = button->y_pos;
     bog_data->w = object->width;
     bog_data->h = object->height;
+    bog_data->visible_object_id = object->id;
 
     gc->ig_drawn = 1;
     gc->ig_dirty = 1;
 }
 
-static void _render_page(GRAPHICS_CONTROLLER *gc,
+static int _render_page(GRAPHICS_CONTROLLER *gc,
                          unsigned activated_button_id,
                          GC_NAV_CMDS *cmds)
 {
@@ -680,30 +755,40 @@ static void _render_page(GRAPHICS_CONTROLLER *gc,
     unsigned        selected_button_id = bd_psr_read(gc->regs, PSR_SELECTED_BUTTON_ID);
 
     if (s->ics->interactive_composition.ui_model == IG_UI_MODEL_POPUP && !gc->popup_visible) {
-        GC_TRACE("_render_page(): popup menu not visible\n");
 
-        _clear_osd(gc, BD_OVERLAY_IG);
+        if (gc->ig_open) {
+            GC_TRACE("_render_page(): popup menu not visible\n");
+            _close_osd(gc, BD_OVERLAY_IG);
+            return 1;
+        }
 
-        return;
+        return 0;
     }
 
     page = _find_page(&s->ics->interactive_composition, page_id);
     if (!page) {
         GC_ERROR("_render_page: unknown page id %d (have %d pages)\n",
               page_id, s->ics->interactive_composition.num_pages);
-        return;
+        return -1;
     }
 
     palette = _find_palette(s, page->palette_id_ref);
     if (!palette) {
         GC_ERROR("_render_page: unknown palette id %d (have %d palettes)\n",
               page->palette_id_ref, s->num_palette);
-        return;
+        return -1;
     }
 
     GC_TRACE("rendering page #%d using palette #%d. page has %d bogs\n",
           page->id, page->palette_id_ref, page->num_bogs);
 
+    if (!gc->ig_open) {
+        _open_osd(gc, BD_OVERLAY_IG,
+                  s->ics->video_descriptor.video_width,
+                  s->ics->video_descriptor.video_height);
+    }
+
+
     for (ii = 0; ii < page->num_bogs; ii++) {
         BD_IG_BOG    *bog      = &page->bog[ii];
         unsigned      valid_id = gc->bog_data[ii].enabled_button;
@@ -751,7 +836,10 @@ static void _render_page(GRAPHICS_CONTROLLER *gc,
     if (gc->ig_dirty) {
         _flush_osd(gc, BD_OVERLAY_IG, -1);
         gc->ig_dirty = 0;
+        return 1;
     }
+
+    return 0;
 }
 
 /*
@@ -1057,6 +1145,11 @@ static int _mouse_move(GRAPHICS_CONTROLLER *gc, unsigned x, unsigned y, GC_NAV_C
         }
 
         new_btn_id = button->id;
+
+        if (cmds) {
+            cmds->sound_id_ref = button->selected_sound_id_ref;
+        }
+
         break;
     }
 
@@ -1077,6 +1170,7 @@ int gc_run(GRAPHICS_CONTROLLER *gc, gc_ctrl_e ctrl, uint32_t param, GC_NAV_CMDS
         cmds->num_nav_cmds = 0;
         cmds->nav_cmds     = NULL;
         cmds->sound_id_ref = -1;
+        cmds->status       = GC_STATUS_NONE;
     }
 
     if (!gc) {
@@ -1133,7 +1227,7 @@ int gc_run(GRAPHICS_CONTROLLER *gc, gc_ctrl_e ctrl, uint32_t param, GC_NAV_CMDS
             /* fall thru */
 
         case GC_CTRL_NOP:
-            _render_page(gc, 0xffff, cmds);
+            result = _render_page(gc, 0xffff, cmds);
             break;
 
         case GC_CTRL_INIT_MENU:
@@ -1163,6 +1257,15 @@ int gc_run(GRAPHICS_CONTROLLER *gc, gc_ctrl_e ctrl, uint32_t param, GC_NAV_CMDS
             break;
     }
 
+    if (cmds) {
+        if (gc->igs->ics->interactive_composition.ui_model == IG_UI_MODEL_POPUP) {
+            cmds->status |= GC_STATUS_POPUP;
+        }
+        if (gc->ig_drawn) {
+            cmds->status |= GC_STATUS_MENU_OPEN;
+        }
+    }
+
     bd_mutex_unlock(&gc->mutex);
 
     return result;
